@@ -10,6 +10,7 @@ using namespace gpd;
 using namespace std;
 using namespace upb;
 using namespace upb::pb;
+using namespace upb::json;
 
 #ifdef MULTIPLICITY
 #    define THX_DECLARE_AND_GET tTHX my_perl = cxt->my_perl
@@ -413,7 +414,8 @@ Mapper::Mapper(pTHX_ Dynamic *_registry, const MessageDef *_message_def, HV *_st
 
     env.ReportErrorsTo(&status);
     registry->ref();
-    encoder_handlers = Encoder::NewHandlers(message_def);
+    pb_encoder_handlers = Encoder::NewHandlers(message_def);
+    json_encoder_handlers = Printer::NewHandlers(message_def, false /* XXX option */);
     decoder_handlers = Handlers::New(message_def);
     decode_explicit_defaults = options.explicit_defaults;
     encode_defaults = options.encode_defaults;
@@ -438,14 +440,17 @@ Mapper::Mapper(pTHX_ Dynamic *_registry, const MessageDef *_message_def, HV *_st
 
             field.name = newSVpvn_share(temp.data(), temp.size(), 0);
         } else {
-            field.name = newSVpvn_share(field_def->name(), strlen(field_def->name()), 0);
+            string temp = string(field_def->name());
+
+            field.name = newSVpvn_share(temp.data(), temp.size(), 0);
         }
         field.name_hash = SvSHARED_HASH(field.name);
+        field.has_default = false;
         field.mapper = NULL;
         field.oneof_index = -1;
 
 #define GET_SELECTOR(KIND, TO) \
-    ok = ok && encoder_handlers->GetSelector(field_def, UPB_HANDLER_##KIND, &field.selector.TO)
+    ok = ok && pb_encoder_handlers->GetSelector(field_def, UPB_HANDLER_##KIND, &field.selector.TO)
 
 #define SET_VALUE_HANDLER(TYPE, FUNCTION) \
     ok = ok && decoder_handlers->SetValueHandler<TYPE>(field_def, UpbBind(DecoderHandlers::FUNCTION, new int(index)))
@@ -552,6 +557,7 @@ Mapper::Mapper(pTHX_ Dynamic *_registry, const MessageDef *_message_def, HV *_st
             extension_mapper_fields.push_back(new MapperField(aTHX_ this, &*it));
             unref(); // to avoid ref loop
         }
+        field_map[SvPV_nolen(it->name)] = &*it;
     }
 
     int oneof_index = 0;
@@ -611,6 +617,36 @@ SV *Mapper::message_descriptor() const {
     return ref;
 }
 
+SV *Mapper::make_object(SV *data) const {
+    SV *obj = NULL;
+
+    if (data) {
+        if (!SvROK(data) || SvTYPE(SvRV(data)) != SVt_PVHV)
+            croak("Not an hash reference");
+
+        if (SvTEMP(data) && SvREFCNT(data) == 1) {
+            // steal
+            SvREFCNT_inc(data);
+            obj = data;
+        } else {
+            HV * hv = newHV();
+            sv_2mortal((SV *) hv);
+            HV *orig = (HV *) SvRV(data);
+            I32 items = hv_iterinit(orig), keylen;
+            char *key;
+            while (SV *sv = hv_iternextsv(orig, &key, &keylen)) {
+                hv_store(hv, key, keylen, newSVsv(sv), 0);
+            }
+            obj = newRV_inc((SV *) hv);
+        }
+    } else {
+        obj = newRV_noinc((SV *) newHV());
+    }
+    sv_bless(obj, stash);
+
+    return obj;
+}
+
 void Mapper::resolve_mappers() {
     for (vector<Field>::iterator it = fields.begin(), en = fields.end(); it != en; ++it) {
         const FieldDef *field = it->field_def;
@@ -622,19 +658,35 @@ void Mapper::resolve_mappers() {
         decoder_handlers->SetSubHandlers(it->field_def, it->mapper->decoder_handlers.get());
     }
 
-    decoder_method = DecoderMethod::New(DecoderMethodOptions(decoder_handlers.get()));
+    pb_decoder_method = DecoderMethod::New(DecoderMethodOptions(decoder_handlers.get()));
+    json_decoder_method = ParserMethod::New(message_def);
     decoder_sink.Reset(decoder_handlers.get(), &decoder_callbacks);
-    decoder = upb::pb::Decoder::Create(&env, decoder_method.get(), &decoder_sink);
-    encoder = upb::pb::Encoder::Create(&env, encoder_handlers.get(), string_sink.input());
+    pb_decoder = upb::pb::Decoder::Create(&env, pb_decoder_method.get(), &decoder_sink);
+    json_decoder = upb::json::Parser::Create(&env, json_decoder_method.get(), &decoder_sink);
+    pb_encoder = upb::pb::Encoder::Create(&env, pb_encoder_handlers.get(), string_sink.input());
+    json_encoder = upb::json::Printer::Create(&env, json_encoder_handlers.get(), string_sink.input());
 }
 
 SV *Mapper::encode(SV *ref) {
-    if (encoder == NULL)
+    if (pb_encoder == NULL)
         croak("It looks like resolve_references() was not called (and please use map() anyway)");
     status.Clear();
     output_buffer.clear();
     SV *result = NULL;
-    if (encode(encoder, encoder->input(), &status, ref))
+    if (encode(pb_encoder->input(), &status, ref))
+        result = newSVpvn(output_buffer.data(), output_buffer.size());
+    output_buffer.clear();
+
+    return result;
+}
+
+SV *Mapper::encode_json(SV *ref) {
+    if (json_encoder == NULL)
+        croak("It looks like resolve_references() was not called (and please use map() anyway)");
+    status.Clear();
+    output_buffer.clear();
+    SV *result = NULL;
+    if (encode(json_encoder->input(), &status, ref))
         result = newSVpvn(output_buffer.data(), output_buffer.size());
     output_buffer.clear();
 
@@ -642,20 +694,41 @@ SV *Mapper::encode(SV *ref) {
 }
 
 SV *Mapper::decode(const char *buffer, STRLEN bufsize) {
-    if (decoder == NULL)
+    if (pb_decoder == NULL)
         croak("It looks like resolve_references() was not called (and please use map() anyway)");
     status.Clear();
-    decoder->Reset();
+    pb_decoder->Reset();
     decoder_callbacks.prepare(newHV());
 
     SV *result = NULL;
-    if (BufferSource::PutBuffer(buffer, bufsize, decoder->input()))
+    if (BufferSource::PutBuffer(buffer, bufsize, pb_decoder->input()))
         result = sv_bless(newRV_inc(decoder_callbacks.get_target()), stash);
     decoder_callbacks.clear();
 
     return result;
 }
 
+SV *Mapper::decode_json(const char *buffer, STRLEN bufsize) {
+    if (json_decoder == NULL)
+        croak("It looks like resolve_references() was not called (and please use map() anyway)");
+    status.Clear();
+    decoder_callbacks.prepare(newHV());
+
+    SV *result = NULL;
+    if (BufferSource::PutBuffer(buffer, bufsize, json_decoder->input()))
+        result = sv_bless(newRV_inc(decoder_callbacks.get_target()), stash);
+    decoder_callbacks.clear();
+
+    return result;
+}
+
+bool Mapper::check(SV *ref) {
+    if (pb_encoder == NULL)
+        croak("It looks like resolve_references() was not called (and please use map() anyway)");
+    status.Clear();
+    return check(&status, ref);
+}
+
 const char *Mapper::last_error_message() const {
     return !decoder_callbacks.error.empty() ? decoder_callbacks.error.c_str() :
            !status.ok()                     ? status.error_message() :
@@ -821,17 +894,14 @@ namespace {
 }
 
 template<class G, class S>
-bool Mapper::encode_from_array(Encoder *encoder, Sink *sink, Status *status, const Mapper::Field &fd, AV *source) const {
+bool Mapper::encode_from_array(Sink *sink, Status *status, const Mapper::Field &fd, AV *source) const {
     G getter;
     S setter(status);
-    Sink sub, *actual = sink;
+    Sink sub;
     bool packed = upb_fielddef_isprimitive(fd.field_def) && fd.field_def->packed();
 
-    if (packed) {
-        if (!sink->StartSequence(fd.selector.seq_start, &sub))
-            return false;
-        actual = &sub;
-    }
+    if (!sink->StartSequence(fd.selector.seq_start, &sub))
+        return false;
     int size = av_top_index(source) + 1;
 
     for (int i = 0; i < size; ++i) {
@@ -839,15 +909,19 @@ bool Mapper::encode_from_array(Encoder *encoder, Sink *sink, Status *status, con
         if (!item)
             return false;
 
-        if (!setter(aTHX_ actual, fd, getter(aTHX_ *item)))
+        if (!setter(aTHX_ &sub, fd, getter(aTHX_ *item)))
             return false;
     }
 
-    return !packed || sink->EndSequence(fd.selector.seq_end);
+    return sink->EndSequence(fd.selector.seq_end);
 }
 
-bool Mapper::encode_from_message_array(Encoder *encoder, Sink *sink, Status *status, const Mapper::Field &fd, AV *source) const {
+bool Mapper::encode_from_message_array(Sink *sink, Status *status, const Mapper::Field &fd, AV *source) const {
     int size = av_top_index(source) + 1;
+    Sink sub;
+
+    if (!sink->StartSequence(fd.selector.seq_start, &sub))
+        return false;
 
     for (int i = 0; i < size; ++i) {
         SV **item = av_fetch(source, i, 0);
@@ -856,15 +930,15 @@ bool Mapper::encode_from_message_array(Encoder *encoder, Sink *sink, Status *sta
         Sink submsg;
 
         SvGETMAGIC(*item);
-        if (!sink->StartSubMessage(fd.selector.msg_start, &submsg))
+        if (!sub.StartSubMessage(fd.selector.msg_start, &submsg))
             return false;
-        if (!encode(encoder, &submsg, status, *item))
+        if (!encode(&submsg, status, *item))
             return false;
-        if (!sink->EndSubMessage(fd.selector.msg_end))
+        if (!sub.EndSubMessage(fd.selector.msg_end))
             return false;
     }
 
-    return true;
+    return sink->EndSequence(fd.selector.seq_end);
 }
 
 namespace {
@@ -876,7 +950,7 @@ namespace {
     }
 }
 
-bool Mapper::encode(Encoder* encoder, Sink *sink, Status *status, SV *ref) const {
+bool Mapper::encode(Sink *sink, Status *status, SV *ref) const {
     if (!SvROK(ref) || SvTYPE(SvRV(ref)) != SVt_PVHV)
         croak("Not an hash reference when encoding a %s value", message_def->full_name());
     HV *hv = (HV *) SvRV(ref);
@@ -907,9 +981,9 @@ bool Mapper::encode(Encoder* encoder, Sink *sink, Status *status, SV *ref) const
         }
 
         if (it->field_def->label() == UPB_LABEL_REPEATED)
-            ok = ok && encode_from_perl_array(encoder, sink, status, *it, HeVAL(he));
+            ok = ok && encode_from_perl_array(sink, status, *it, HeVAL(he));
         else
-            ok = ok && encode(encoder, sink, status, *it, HeVAL(he));
+            ok = ok && encode(sink, status, *it, HeVAL(he));
     }
 
     if (!sink->EndMessage(status))
@@ -918,7 +992,7 @@ bool Mapper::encode(Encoder* encoder, Sink *sink, Status *status, SV *ref) const
     return ok;
 }
 
-bool Mapper::encode(Encoder* encoder, Sink *sink, Status *status, const Field &fd, SV *ref) const {
+bool Mapper::encode(Sink *sink, Status *status, const Field &fd, SV *ref) const {
     switch (fd.field_def->type()) {
     case UPB_TYPE_FLOAT:
         return sink->PutFloat(fd.selector.primitive, SvNV(ref));
@@ -940,7 +1014,7 @@ bool Mapper::encode(Encoder* encoder, Sink *sink, Status *status, const Field &f
         Sink sub;
         if (!sink->StartSubMessage(fd.selector.msg_start, &sub))
             return false;
-        if (!fd.mapper->encode(encoder, &sub, status, ref))
+        if (!fd.mapper->encode(&sub, status, ref))
             return false;
         return sink->EndSubMessage(fd.selector.msg_end);
     }
@@ -977,48 +1051,166 @@ bool Mapper::encode(Encoder* encoder, Sink *sink, Status *status, const Field &f
     }
 }
 
-bool Mapper::encode_from_perl_array(Encoder* encoder, Sink *sink, Status *status, const Field &fd, SV *ref) const {
+bool Mapper::encode_from_perl_array(Sink *sink, Status *status, const Field &fd, SV *ref) const {
     if (!SvROK(ref) || SvTYPE(SvRV(ref)) != SVt_PVAV)
         croak("Not an array reference when encoding field '%s'", fd.full_name().c_str());
     AV *array = (AV *) SvRV(ref);
 
     switch (fd.field_def->type()) {
     case UPB_TYPE_FLOAT:
-        return encode_from_array<NVGetter, FloatEmitter>(encoder, sink, fd, array);
+        return encode_from_array<NVGetter, FloatEmitter>(sink, fd, array);
     case UPB_TYPE_DOUBLE:
-        return encode_from_array<NVGetter, DoubleEmitter>(encoder, sink, fd, array);
+        return encode_from_array<NVGetter, DoubleEmitter>(sink, fd, array);
     case UPB_TYPE_BOOL:
-        return encode_from_array<BoolGetter, BoolEmitter>(encoder, sink, fd, array);
+        return encode_from_array<BoolGetter, BoolEmitter>(sink, fd, array);
     case UPB_TYPE_STRING:
-        return encode_from_array<SVGetter, StringEmitter>(encoder, sink, fd, array);
+        return encode_from_array<SVGetter, StringEmitter>(sink, fd, array);
     case UPB_TYPE_BYTES:
-        return encode_from_array<SVGetter, StringEmitter>(encoder, sink, fd, array);
+        return encode_from_array<SVGetter, StringEmitter>(sink, fd, array);
     case UPB_TYPE_MESSAGE:
-        return fd.mapper->encode_from_message_array(encoder, sink, status, fd, array);
+        return fd.mapper->encode_from_message_array(sink, status, fd, array);
     case UPB_TYPE_ENUM:
         if (check_enum_values)
-            return encode_from_array<IVGetter, EnumEmitter>(encoder, sink, status, fd, array);
+            return encode_from_array<IVGetter, EnumEmitter>(sink, status, fd, array);
         else
-            return encode_from_array<IVGetter, Int32Emitter>(encoder, sink, fd, array);
+            return encode_from_array<IVGetter, Int32Emitter>(sink, fd, array);
     case UPB_TYPE_INT32:
-        return encode_from_array<IVGetter, Int32Emitter>(encoder, sink, fd, array);
+        return encode_from_array<IVGetter, Int32Emitter>(sink, fd, array);
     case UPB_TYPE_UINT32:
-        return encode_from_array<UVGetter, UInt32Emitter>(encoder, sink, fd, array);
+        return encode_from_array<UVGetter, UInt32Emitter>(sink, fd, array);
     case UPB_TYPE_INT64:
         if (sizeof(IV) >= sizeof(int64_t))
-            return encode_from_array<IVGetter, Int64Emitter>(encoder, sink, fd, array);
+            return encode_from_array<IVGetter, Int64Emitter>(sink, fd, array);
         else
-            return encode_from_array<I64Getter, Int64Emitter>(encoder, sink, fd, array);
+            return encode_from_array<I64Getter, Int64Emitter>(sink, fd, array);
     case UPB_TYPE_UINT64:
         if (sizeof(IV) >= sizeof(int64_t))
-            return encode_from_array<UVGetter, UInt64Emitter>(encoder, sink, fd, array);
+            return encode_from_array<UVGetter, UInt64Emitter>(sink, fd, array);
         else
-            return encode_from_array<U64Getter, UInt64Emitter>(encoder, sink, fd, array);
+            return encode_from_array<U64Getter, UInt64Emitter>(sink, fd, array);
     default:
         return false; // just in case
     }
 }
 
+bool Mapper::check_from_message_array(Status *status, const Mapper::Field &fd, AV *source) const {
+    int size = av_top_index(source) + 1;
+
+    for (int i = 0; i < size; ++i) {
+        SV **item = av_fetch(source, i, 0);
+        if (!item)
+            return false;
+
+        SvGETMAGIC(*item);
+        if (!check(status, *item))
+            return false;
+    }
+
+    return true;
+}
+
+bool Mapper::check_from_enum_array(Status *status, const Mapper::Field &fd, AV *source) const {
+    int size = av_top_index(source) + 1;
+
+    for (int i = 0; i < size; ++i) {
+        SV **item = av_fetch(source, i, 0);
+        if (!item)
+            return false;
+
+        IV value = SvIV(*item);
+        if (fd.enum_values.find(value) == fd.enum_values.end()) {
+            status->SetFormattedErrorMessage(
+                "Invalid enumeration value %d for field '%s'",
+                value,
+                fd.full_name().c_str()
+            );
+            return false;
+        }
+    }
+
+    return true;
+}
+
+bool Mapper::check(Status *status, SV *ref) const {
+    if (!SvROK(ref) || SvTYPE(SvRV(ref)) != SVt_PVHV)
+        croak("Not an hash reference when checking a %s value", message_def->full_name());
+    HV *hv = (HV *) SvRV(ref);
+
+    I32 count = hv_iterinit(hv);
+    bool ok = true;
+    for (int i = 0; i < count; ++i) {
+        char *key;
+        I32 keylen;
+        SV *value = hv_iternextsv(hv, &key, &keylen);
+        // if the key is marked as UTF-8 and contains non-ASCII characters,
+        // it will not be there anyway in the lookup
+        string name(key, keylen < 0 ? -keylen : keylen);
+        STD_TR1::unordered_map<string, Field *>::const_iterator it = field_map.find(name);
+
+        if (it == field_map.end()) {
+            status->SetFormattedErrorMessage(
+                "Unknown field '%s' during check",
+                name.c_str());
+            return false;
+        }
+
+        Field *field = it->second;
+        if (field->field_def->label() == UPB_LABEL_REPEATED)
+            ok = ok && check_from_perl_array(status, *field, value);
+        else
+            ok = ok && check(status, *field, value);
+    }
+
+    return ok;
+}
+
+bool Mapper::check(Status *status, const Field &fd, SV *ref) const {
+    switch (fd.field_def->type()) {
+    case UPB_TYPE_MESSAGE:
+        return fd.mapper->check(status, ref);
+    case UPB_TYPE_ENUM: {
+        if (!check_enum_values)
+            return true;
+
+        IV value = SvIV(ref);
+        if (fd.enum_values.find(value) == fd.enum_values.end()) {
+            status->SetFormattedErrorMessage(
+                "Invalid enumeration value %d for field '%s'",
+                value,
+                fd.full_name().c_str()
+            );
+            return false;
+        }
+
+        return true;
+    }
+    default:
+        // I doubt there is any point in performing "strict" type checks
+        // on scalar values, due to coercion
+        return true;
+    }
+}
+
+bool Mapper::check_from_perl_array(Status *status, const Field &fd, SV *ref) const {
+    if (!SvROK(ref) || SvTYPE(SvRV(ref)) != SVt_PVAV)
+        croak("Not an array reference when encoding field '%s'", fd.full_name().c_str());
+    AV *array = (AV *) SvRV(ref);
+
+    switch (fd.field_def->type()) {
+    case UPB_TYPE_MESSAGE:
+        return fd.mapper->check_from_message_array(status, fd, array);
+    case UPB_TYPE_ENUM:
+        if (check_enum_values)
+            return check_from_enum_array(status, fd, array);
+        else
+            return true;
+    default:
+        // I doubt there is any point in performing "strict" type checks
+        // on scalar values, due to coercion
+        return true;
+    }
+}
+
 MapperField::MapperField(pTHX_ const Mapper *_mapper, const Mapper::Field *_field) :
         field(_field),
         mapper(_mapper) {
@@ -1373,3 +1565,24 @@ void MapperField::set_list(HV *self, SV *ref) {
     SvRV_set(field_ref, SvRV(ref));
     SvREFCNT_inc(SvRV(field_ref));
 }
+
+EnumMapper::EnumMapper(pTHX_ Dynamic *_registry, const upb::EnumDef *_enum_def) :
+        registry(_registry),
+        enum_def(_enum_def) {
+    SET_THX_MEMBER;
+
+    registry->ref();
+}
+
+EnumMapper::~EnumMapper() {
+    // make sure this only goes away after inner destructors have completed
+    refcounted_mortalize(aTHX_ registry);
+}
+
+SV *EnumMapper::enum_descriptor() const {
+    SV *ref = newSV(0);
+
+    sv_setref_iv(ref, "Google::ProtocolBuffers::Dynamic::EnumDef", (IV) enum_def);
+
+    return ref;
+}
