@@ -7,6 +7,8 @@
 
 #include <upb/pb/encoder.h>
 #include <upb/pb/decoder.h>
+#include <upb/json/printer.h>
+#include <upb/json/parser.h>
 #include <upb/bindings/stdc++/string.h>
 
 #include "unordered_map.h"
@@ -106,6 +108,9 @@ public:
 
     SV *encode(SV *ref);
     SV *decode(const char *buffer, STRLEN bufsize);
+    SV *encode_json(SV *ref);
+    SV *decode_json(const char *buffer, STRLEN bufsize);
+    bool check(SV *ref);
 
     const char *last_error_message() const;
 
@@ -115,38 +120,49 @@ public:
     MapperField *find_extension(const std::string &name) const;
 
     SV *message_descriptor() const;
+    SV *make_object(SV *data) const;
 
 private:
-    bool encode(upb::pb::Encoder* encoder, upb::Sink *sink, upb::Status *status, SV *ref) const;
-    bool encode(upb::pb::Encoder* encoder, upb::Sink *sink, upb::Status *status, const Field &fd, SV *ref) const;
-    bool encode_from_perl_array(upb::pb::Encoder* encoder, upb::Sink *sink, upb::Status *status, const Field &fd, SV *ref) const;
-    bool encode_from_message_array(upb::pb::Encoder *encoder, upb::Sink *sink, upb::Status *status, const Mapper::Field &fd, AV *source) const;
+    bool encode(upb::Sink *sink, upb::Status *status, SV *ref) const;
+    bool encode(upb::Sink *sink, upb::Status *status, const Field &fd, SV *ref) const;
+    bool encode_from_perl_array(upb::Sink *sink, upb::Status *status, const Field &fd, SV *ref) const;
+    bool encode_from_message_array(upb::Sink *sink, upb::Status *status, const Mapper::Field &fd, AV *source) const;
 
     template<class G, class S>
-    bool encode_from_array(upb::pb::Encoder *encoder, upb::Sink *sink, upb::Status *status, const Mapper::Field &fd, AV *source) const;
+    bool encode_from_array(upb::Sink *sink, upb::Status *status, const Mapper::Field &fd, AV *source) const;
 
     template<class G, class S>
-    bool encode_from_array(upb::pb::Encoder *encoder, upb::Sink *sink, const Mapper::Field &fd, AV *source) const {
-        return encode_from_array<G, S>(encoder, sink, NULL, fd, source);
+    bool encode_from_array(upb::Sink *sink, const Mapper::Field &fd, AV *source) const {
+        return encode_from_array<G, S>(sink, NULL, fd, source);
     }
 
+    bool check(upb::Status *status, SV *ref) const;
+    bool check(upb::Status *status, const Field &fd, SV *ref) const;
+    bool check_from_perl_array(upb::Status *status, const Field &fd, SV *ref) const;
+    bool check_from_message_array(upb::Status *status, const Mapper::Field &fd, AV *source) const;
+    bool check_from_enum_array(upb::Status *status, const Mapper::Field &fd, AV *source) const;
+
     DECL_THX_MEMBER;
     Dynamic *registry;
     const upb::MessageDef *message_def;
     HV *stash;
-    upb::reffed_ptr<const upb::Handlers> encoder_handlers;
+    upb::reffed_ptr<const upb::Handlers> pb_encoder_handlers, json_encoder_handlers;
     upb::reffed_ptr<upb::Handlers> decoder_handlers;
-    upb::reffed_ptr<const upb::pb::DecoderMethod> decoder_method;
+    upb::reffed_ptr<const upb::pb::DecoderMethod> pb_decoder_method;
+    upb::reffed_ptr<const upb::json::ParserMethod> json_decoder_method;
     std::vector<Field> fields;
     std::vector<MapperField *> extension_mapper_fields;
+    STD_TR1::unordered_map<std::string, Field *> field_map;
     upb::Environment env;
     upb::Status status;
     DecoderHandlers decoder_callbacks;
     upb::Sink encoder_sink, decoder_sink;
-    upb::pb::Decoder *decoder;
+    upb::pb::Decoder *pb_decoder;
+    upb::json::Parser *json_decoder;
     std::string output_buffer;
     upb::StringSink string_sink;
-    upb::pb::Encoder *encoder;
+    upb::pb::Encoder *pb_encoder;
+    upb::json::Printer *json_encoder;
     bool check_required_fields, decode_explicit_defaults, encode_defaults, check_enum_values;
 };
 
@@ -194,6 +210,19 @@ private:
     const Mapper *mapper;
 };
 
+class EnumMapper : public Refcounted {
+public:
+    EnumMapper(pTHX_ Dynamic *registry, const upb::EnumDef *enum_def);
+    ~EnumMapper();
+
+    SV *enum_descriptor() const;
+
+private:
+    DECL_THX_MEMBER;
+    Dynamic *registry;
+    const upb::EnumDef *enum_def;
+};
+
 };
 
 #endif
